!<CPS_v1 type="PROGRAM"/>
!!------------------------------ segyview.f90 --------------------------------!!
!!------------------------------ segyview.f90 --------------------------------!!
!!------------------------------ segyview.f90 --------------------------------!!


!<license>
!-------------------------------------------------------------------------------
! Copyright (c) 2007 ConocoPhillips Company
!
! Permission is hereby granted, free of charge, to any person obtaining a copy
! of this software and associated documentation files (the "Software"), to deal
! in the Software without restriction, including without limitation the rights
! to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
! copies of the Software, and to permit persons to whom the Software is
! furnished to do so, subject to the following conditions:
!
! The above copyright notice and this permission notice shall be included in all
! copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
! IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
! FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
! AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
! LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
! SOFTWARE.
!-------------------------------------------------------------------------------
!</license>


!<brief_doc>
!-------------------------------------------------------------------------------
!                       C P S   P R O G R A M 
!
! Name       : segyview
! Category   : stand-alone
! Written    : 2001-01-03   by: Bill Menger
! Revised    : 2001-09-07   by: Bill Menger
! Revised    : 2017-July
! Maturity   : beta
! Purpose    : View SEGY files using TEXT-only quick viewer
! Portability: No known limitations.
!
!-------------------------------------------------------------------------------
!</brief_doc>


!<descript_doc>
!-------------------------------------------------------------------------------
!                          GENERAL DESCRIPTION
! SEGYVIEW uses the base functions of CPS to open, read, and display the 
! contents from SEGY files.  One may view the EBCDIC header, BINARY header, and
! any trace header or trace values (using a tab-plot).  If a SEGY file cannot be
! opened with SEGYVIEW then it cannot be read by CPS.
!-------------------------------------------------------------------------------
!</descript_doc>


!<advice_doc>
!-------------------------------------------------------------------------------
!                          ADVICE FOR USERS
! Traces are read as CPSeis then headers converted to SEG Y, so not all SEG Y
! headers will be printed.
!-----------------------------------12345678901234--------------------------------------------
!</advice_doc>


!<history_doc>
!-------------------------------------------------------------------------------
!                             REVISION HISTORY
!
!     Date        Author      Description
!     ----        ------      -----------
!  4. 2017-02-31		added partial revision 1 handling
!  3. 2017-07-29		fixed 2 bugs
!  2. 2001-09-07  Bill Menger Modified to use the put module,added ident.
!  1. 2001-01-03  Bill Menger Initial version.
!-------------------------------------------------------------------------------
!</history_doc>


!<portability_doc>
!-------------------------------------------------------------------------------
!                         PORTABILITY LIMITATIONS
!
! No known limitations.
!-------------------------------------------------------------------------------
!</portability_doc>

!!-------------------------- start of program ------------------------------!!
!!-------------------------- start of program ------------------------------!!
!!-------------------------- start of program ------------------------------!!

program segyview
use put_module
use named_constants_module
use wrdc_module
use swap_module
use cio_module
use segy_module
use trcio_module
implicit none

character(len=100) :: segyview_ident = &
'$Id: segyview.f90,v 1.2 2008/02/15 20:18:51 mengewm Exp $'
type(trcio_struct),pointer    :: file
type(segy_ebcdic_hdr) :: ebcdic
type(segy_bin_hdr)    :: binhdr
type(segy_trc_hdr)    :: segyhd
double precision      :: cpshd(64)
real,allocatable      :: trace(:)
integer               :: h(60) , err
integer		      :: datastart(2)	!!! added 2017
character(len=8)      :: key
character(len=128)    :: filename
logical               :: swap,ibmswap,newtrace,phlush
double precision      :: dt,tstart
integer               :: ns,trcnum,neth		!!! added neth
character(len=1)      :: go,cont
character(len=32)     :: tracenumber,values
character(len=132)    :: str
character(len=12)     :: stanza			!!! added 2017

integer :: i,j,unit,status,ifirst,ilast,irange

	write(6,*) '    [text-mode segyview]'
allocate(file)				! added 2017
100 continue
if(associated(file)) status = trcio_close(file)

status=put('Segy File (q to stop): ')
read(5,'(A)')filename
status=put('filename: %s'//char(10),trim(filename))

if(trim(filename) == 'q' .or. trim(filename) == 'Q') stop
if(len_trim(filename) <= 0 ) go to 100
file => trcio_open(filename,'r')	! this prints binary header
if(.not. associated(file)) then
  status=put('Error opening '//trim(filename)//'.  Read another?(y/n): ')
  read*,cont
  if(cont == 'y' .or. cont == 'Y') then
    go to 100
  else
    stop
  endif
endif
unit = file%lun
call cio_frewind(unit)

if(segy_is_file_segy(unit) == 1)  then
  status=put(trim(filename)//' is not SEGY.  Read another?(y/n): ')
  read*,cont
  if(cont == 'y' .or. cont == 'Y') then
    go to 100
  else
    stop
  endif
endif	! is segy
	call cio_frewind(unit)		! added 2017
status = segy_read_ebchdr(unit,ebcdic)	! returns ASCII
if(status /= 0 ) then
  status=&
  put('Error reading ebcdic header.  Status= %d Keep going?(y/n): ',status)
  read*,cont
  if(cont /= 'y' .and. cont /= 'Y') go to 100
endif	! status
do i = 1, 40
  status=put(ebcdic%h(i)//char(10))
end do
!!! read 3200 bytes
!status=put('Hit return to continue')
!read(5,'(a)')cont
status = segy_read_binhdr(unit,binhdr) ! binhdr is type(segy_bin_hdr)
if(status /= 0 ) then
  status=&
  put('Error reading binary header.  Status= %d Keep going?(y/n): ',status)
  read*,cont
  if(cont /= 'y' .and. cont /= 'Y') go to 100
endif ! status
!!! read 3600 bytes so far...
! 2017 suppressed this, because already done: 
!call segy_print_binhdr(binhdr)

!if (swap_endian() /= binhdr%endian ) then 
!  swap = .true.
!else
!  swap = .false.
!endif
!!! get extra text headers (and case of -1):
if (binhdr%netfh .lt. 0)  binhdr%netfh=999
if (binhdr%netfh .gt. 0) then
 do neth = 1,binhdr%netfh
  write(6,*) '---Extra text headers---',neth
  status = segy_read_ebchdr(unit,ebcdic)	! returns ASCII
  if(status /= 0) write(6,*) '[error reading extra headers]'
  do i = 1, 40
   status=put(ebcdic%h(i)//char(10))
  end do
  stanza=ebcdic%h(1)(3:14)
  if((stanza .eq. 'SEG: EndText').or.(stanza .eq. 'SEG: ENDTEXT')) exit
 end do
! if end loop normally, neth=binhdr%netfh+1
! if exit, neth=actual number of extra headers
! write(6,*) 'n ',neth,binhdr%netfh
 if (binhdr%netfh .eq. 999) binhdr%netfh=neth
 datastart(1)=0
 datastart(2)=3600+binhdr%netfh*3200
 status= trcio_set_data_start_pos(file,datastart)
endif ! extra headers
!!! trcio_set_data_end_pos seems not to be needed

allocate(trace(int(binhdr%hns)))
trcnum = 0
do
  trcnum = trcnum + 1
  status=put('Trace number to read[%d]:',trcnum)
  read(5,'(a)')tracenumber
  newtrace=.true.
  if(len_trim(tracenumber) == 0 ) then ! nothing?
  else
    if(trim(tracenumber) == 'q' .or. trim(tracenumber)=='Q') exit
    read(tracenumber,*)trcnum
  endif
! trcio_read_trace calls segy_unpack_segyhd,segy_segyhd_to_cpshd
  status = trcio_read_trace(file,cpshd,trace,trcnum)
  if(status /= 0 ) then
    status=&
    put('Error reading trace.  Status= %d Keep going?(y/n): ',status)
    read*,cont
    if(cont /= 'y' .and. cont /= 'Y') exit
  endif ! status
  write(6,178) cpshd(7),cpshd(8),cpshd(17),cpshd(18),cpshd(29)	!!! debug
178   format(F9.1,F9.1,F9.1,F9.1,F9.1)
  tstart = file%tmin
  dt     = file%dt
  ns     = file%num_values
! convert headers to SEG Y headers (as would be done for output)
  call segy_cpshd_to_segyhd(cpshd,segyhd,tstart,dt,ns)
! headers have been converted back and forth by now!
  call segy_print_trace_hd(segyhd)
  phlush=.false.
!  call segy_hdrdump(phlush,segyhd)
  phlush=.true.
!  call segy_hdrdump(phlush)
  irange = 20
  ifirst = 1          
  ilast  = min(irange,nint(cpshd(HDR_BOTTOM_MUTE))-1)
200 continue
    ifirst    = min(int(segyhd%ns),max(1,ifirst))
    if(newtrace) then
      newtrace=.false.
    else
      ilast     = max(ifirst,min(ilast,int(segyhd%ns)))
    endif	! newtrace
    status=put('Samples to plot[%d] [',ifirst)
    status=put('%d]:',ilast)
    read(5,'(a)')values
    if(len_trim(values) /= 0 ) then 
      if(trim(values)=='q' .or. trim(values)=='Q' .or. trim(values) =='0')cycle
      read(values,*,err=200)ifirst,ilast
    endif
  ifirst    = min(int(segyhd%ns),max(1,ifirst))
  ilast     = max(ifirst,min(ilast,int(segyhd%ns)))
  call segy_tabplot(trace(1),ifirst,ilast)
  irange = ilast - ifirst + 1
  ifirst = ilast + 1
  ilast  = ifirst + irange - 1
  go to 200
end do
deallocate (trace)
go to 100

end program segyview

!!-------------------------------- end -----------------------------------!!
!!-------------------------------- end -----------------------------------!!
!!-------------------------------- end -----------------------------------!!
